//---------------------------------------------------------------------------

#include "stdafx.h"

#ifdef _MSC_VER
#pragma hdrstop
#endif

using namespace std;

#include "../../Kernel/kernelincludes.h"

#include "defaults.h"
//#include "TimeStamp.h"
#include "Expand.h"
//---------------------------------------------------------------------------

Expand::Expand( PLA *_source, PLA *_implpool )
{
    implpool = _implpool;
    source = _source;
    method = DEFAULT_IE;
}

Expand::~Expand(void)
{
}

void Expand::FastExpand(void)
//	Starting from a random position removes literals, while the term is an implicant of ov.
// All output variables are handled separately, thus the number of implicants may be increased.
// Keeps unexpanded implicants in the pool
{
    unsigned int ov;
	unsigned int i, i_start;
    TermList::iterator ti;
    TermList newterms;          // newly created terms
    Term t;                     // term under construction
    TermList::iterator ti1;     // newly created term
    bool change;                // is there any change?
    char c;

	for ( ov = 0; ov < implpool->outputs(); ov++ ) {
		for ( ti = implpool->getData()->begin(); ti != implpool->getData()->end(); ti++ )
		  if ( (*ti).isOM_1(ov) ) {
			t = *ti;
			change = false;
			i_start = rand() % implpool->inputs();
			i = i_start;
			do {
				if ( t.isLit_dc(i) ) {
					c = t.getLit(i);
					t.setLit_dc(i);
					if ( !source->isImplicant( t, ov ) ) t.setLit( i, c );
					else change = true;
				}
				i++;
				i %= implpool->inputs();
			} while ( i != i_start );
			if ( change ) {
				for ( i = 0; i < t.getOMSize(); i++ )
					if ( i != ov && t.isOM_1(i) )
						if ( !source->isImplicant( t, i ) ) t.setOM_0(i);
				t.origin |= 4;
				newterms.push_back(t);
			}
		}
		for ( ti = newterms.begin(); ti != newterms.end(); ti++ ) {
			ti1 = implpool->AppendTerm(*ti);
			for ( i = 0; i < (*ti).getOMSize(); i++ )                    // for a case the term is already present in TT, but having different OM
				if ( (*ti1).isOM_undef(i) && !(*ti).isOM_undef(i) )
					(*ti1).setOM(i, (*ti).getOM(i));
		}
	}
}

void Expand::MultipleExpand(void)
// Removes literals while the term is implicant of ov. Starts independently from all positions.
// Keeps unexpanded implicants in the pool
{
    unsigned int ov;
	unsigned int i;
    TermList::iterator ti;
    Term t;                     // term under construction
    TermList::iterator ti1;                   // newly created term
	TermList::iterator ti_end;                   // end of pool
	unsigned int inputs;		// local copy
	list<Term> *data;					// local copy
    char c;

	inputs = implpool->inputs();
	data = implpool->getData();
	ti_end = data->end();
	for ( ov = 0; ov < implpool->outputs(); ov++ ) {
		for ( ti = data->begin(); ti != ti_end; ++ti )
		  if ( ti->isOM_1(ov) ) {
			t = *ti;
			for ( i = 0; i < inputs; i++ )
				if ( !t.isLit_dc(i) ) {
					c = t.getLit(i);
					t.setLit_dc(i);
					if ( !source->isImplicant( t, ov ) ) t.setLit( i, c );
					else {
						for ( i = 0; i < t.getOMSize(); i++ )
							if ( i != ov && t.isOM_1(i) )
								if ( !source->isImplicant( t, i ) ) t.setOM_0(i);
						t.origin |= 4;
						ti1 = implpool->AppendTerm(t);
						for ( i = 0; i < t.getOMSize(); i++ )                    // for a case the term is already present in TT, but having different OM
							if ( ti1->isOM_undef(i) && !ti->isOM_undef(i) )
								ti1->setOM(i, t.getOM(i));
					}
				}
		}
	}
}

void Expand::ExhaustiveExpand(void)
//	Generates all prime implicants of ov
//  Keeps unexpanded and partially expanded implicants in the pool
{
    unsigned int ov;
	unsigned int i, j;
    TermList::iterator ti;
    Term t;                     // term under construction
    TermList::iterator ti1;     // newly created term
    char c;

	for ( ov = 0; ov < implpool->outputs(); ov++ ) {
		for ( ti = implpool->getData()->begin(); ti != implpool->getData()->end(); ti++ )
		  if ( (*ti).isOM_1(ov) ) {
			t = *ti;
			for ( i = 0; i < implpool->inputs(); i++ )
				if ( !t.isLit_dc(i) ) {
					c = t.getLit(i);
					t.setLit_dc(i);
					if ( !source->isImplicant( t, ov ) ) t.setLit( i, c );
					else {
						for ( j = 0; j < t.getOMSize(); j++ )
							if ( j != ov && t.isOM_1(j) )
								if ( !source->isImplicant( t, j ) ) t.setOM_0(j);
						t.origin |= 4;
						ti1 = implpool->AppendTerm(t);
						for ( j = 0; j < t.getOMSize(); j++ )                    // for a case the term is already present in TT, but having different OM
							if ( (*ti1).isOM_undef(j) && !(*ti).isOM_undef(j) )
								(*ti1).setOM(j, t.getOM(j));
						t.setLit( i, c );
					}
				}
		}
	}
}

void Expand::ScatterExpand(void)
// Randomly removes literals from term, so that no 1 in OM is harmed. Does not change the number of terms or their location
// PLA must not be set "unique"!!! It changes terms directly.
{
	unsigned int pos;			// random term & position
	unsigned int termpos;		// random position in the term
    TermList::iterator ti;		// processed term
	char c;					// removed literal
	vector<bool> tried;			// already tried for removal
	unsigned int trials;		// countdown

	tried.assign( implpool->getTerms() * implpool->inputs(), false );
	trials = implpool->getTerms() * implpool->inputs();

	do {
		pos = rand() % (implpool->getTerms() * implpool->inputs());
		do {
			if ( !tried[pos] ) {
				ti = implpool->getTermByIndex( pos / implpool->inputs() );
				termpos = pos % implpool->inputs();
				tried[pos] = true;
				trials--;
				if ( !(*ti).isLit_dc(termpos) ) {
					c = (*ti).getLit(termpos);
					(*ti).setLit_dc(termpos);
					if ( !source->isImplicant(*ti) ) (*ti).setLit(termpos, c);
					else (*ti).origin |= 4;
				}
				break;
			} else pos = (pos+1) % (implpool->getTerms() * implpool->inputs());
		} while ( trials > 0 );
	} while ( trials > 0 );
}

void Expand::ExpandTerm( Term &t )
// Tries to expand a single term, so that the output matrix will be kept
// Starts from a random position, like "fast"
{
	unsigned int start, i, ov;
	char c;

	start = rand() % source->inputs();
	i = start;
	do {
		c = t.getLit(i);
		if ( c != '-' ) {	
			t.setLit_dc(i);
			for ( ov = 0; ov < source->outputs(); ov++ )
				if ( t.isOM_1(ov) )
					if ( !source->isImplicant( t, ov ) ) break;
			if ( ov != source->outputs() ) t.setLit(i, c);
		}
		i++;
		i %= source->inputs();
	} while ( i != start );
}

void Expand::Run(void)
{
 //   TS.start(TS_IE);

	if ( implpool == NULL ) return;
    switch ( method ) {
		case no_IE:
			break;
        case fast_IE:
            FastExpand();
            break;
        case multiple_IE:
            MultipleExpand();
            break;
        case exhaustive_IE:
            ExhaustiveExpand();
            break;
        case scatter_IE:
            ScatterExpand();
            break;
    }
//	TS.stop(TS_IE);
}
