// GetInfo.cpp : Defines the entry point for the console application.

#include "stdafx.h"

#ifdef _MSC_VER
#pragma hdrstop
#endif

using namespace std;

#include "../../Kernel/kernelincludes.h"
#include "../../Multilevel/Blif.h"
#include "../../Multilevel/Node.h"
#include "../../Multilevel/Signal.h"
#include "../../Multilevel/Network.h"

void help()
{
	cout << "Clean: removes constant outputs from the network" << endl;
	cout << "Syntax: Clean [-wires] source.blif dest.blif" << endl;
	cout << "    -wires   Removes also through i/o connections" << endl;

}

void DeleteConstantOutputs( Network &net )
{
	unsigned int i;
	pair<bool, list<Signal>::iterator> findres;
	pair<list<Node>::iterator, unsigned int> *sigres;

	for ( i = 0; i < net.outputs(); i++ )
		if ( net.isConstant(net.getOName(i)) ) {
			findres = net.FindSignal(net.getOName(i));
			if ( !findres.first ) throw RuntimeException( "DeleteConstantOutputs:: Inconsistency between signals" );
			sigres = findres.second->GetSource();
			net.DeleteNode(sigres->first);
		
		} else {		// try, if it is not fed by a constant node
			findres = net.FindSignal(net.getOName(i));
			if ( !findres.first ) throw RuntimeException( "DeleteConstantOutputs:: Inconsistency between signals" );
			sigres = findres.second->GetSource();
			if ( sigres->first->isBuffer() ) {		// it is a buffer. Try, if its input is not a constant
				findres = net.FindSignal( sigres->first->getConstDataRef().getIName(0) );
				if ( !findres.first ) throw RuntimeException( "DeleteConstantOutputs:: Inconsistency between signals" );
				if ( !findres.second->IsPI() && findres.second->IsConstant() ) {
					net.DeleteNode(sigres->first);
					if ( findres.second->GetDest()->size() == 0 ) net.DeleteNode(findres.second->GetSource()->first);	// the constant is not used any more. Void node.
				}
			}
		}
}

void DeleteThroughConnections( Network &net )
{
	unsigned int i;
	pair<bool, list<Signal>::iterator> findres;
	pair<list<Node>::iterator, unsigned int> *sigres;

	for ( i = 0; i < net.outputs(); i++ ) {
		findres = net.FindSignal(net.getOName(i));
		if ( !findres.first ) throw RuntimeException( "DeleteThroughConnections:: Inconsistency between signals" );
		sigres = findres.second->GetSource();
		if ( sigres->first->isBuffer() ) net.DeleteNode(sigres->first);
	}
}

int main(int argc, char* argv[])
{
	Blif source;
	unsigned int i;
	bool wires = false;
	char *src_file = NULL;
	char *dest_file = NULL;
	Network *net;

	cerr << "Clean 1.0, (c) Petr Fiser, compiled " << __DATE__ << ", " << __TIME__ << endl;
    if ( argc < 3 ) {
		help(); 
		exit(1);
    }

	i = 1;
	while ( i != argc ) {
		if ( argv[i][0] != '-' && src_file == NULL ) src_file = argv[i];
		else if ( argv[i][0] != '-' ) dest_file = argv[i];
		else if ( strcmp( argv[i], "-wires" ) == 0 ) {
			wires = true;
		}
		else {
			help();
			exit(1);
		}
		i++;
	}

	if ( src_file == NULL || dest_file == NULL ) {
        help();
        exit(1);
    }

	source.LoadBlif(src_file);
	net = new Network(source);
	DeleteConstantOutputs(*net);
	net->DeleteVoidPorts();
	if ( wires ) DeleteThroughConnections(*net);
	net->DeleteVoidPorts();
	net->ToBlif()->SaveBlif(dest_file);



	delete net;
    return 0;
}
