// GenLibGen.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <time.h>


using namespace std;

#include "../../Kernel/kernelincludes.h"

void Help()
{
		cerr << "Syntax: GenLibGen -i n [-all | -P | -NP | -NPN] dest.genlib" << endl;
		cerr << "   Options: " << endl;
		cerr << "       -i n     Number of function's inputs" << endl;
		cerr << "       -all     Generates all functions (default)" << endl;
		cerr << "       -P       Generates P classes" << endl;
		cerr << "       -NP      Generates NP classes" << endl;
		cerr << "       -NPN     Generates NPN classes" << endl;
}

void GenerateAll( list<PLA> &functions, unsigned int inputs )
{
	PLA *pla;
	Term *t;
	unsigned int i, j;
	string s;
	list<Term>::iterator ti;

	functions.clear();

		// Generate PLA	of all minterms
	pla = new PLA(inputs, 1, PLA_type("fd"));
	t = new Term(inputs, 1);
	t->setOM_0(0);
	for ( i = 0; i < (1 << inputs); i++ ) {
		t->setCube( int2bin(i, inputs) );
		pla->AppendTerm(*t);
	}

		// Generate 2^2^n functions
	for ( i = 0; i < (1 << (1 << inputs)); i++ ) {
		ti = pla->getData()->begin();
		s = int2bin(i, 1 << inputs);
		for ( j = 0; j < (1 << inputs); j++ ) {
			if ( s[j] == '0' ) ti->setOM_0(0);
			else if ( s[j] == '1' ) ti->setOM_1(0);
			ti++;
		}
		functions.push_back(*pla);
	}
	delete t;
	delete pla;
}

void Minimize( list<PLA> &functions )
{
	list<PLA>::iterator fn_it;
	list<PLA>::iterator fn_it_end;
	PLA pla;

	fn_it_end = functions.end();
	fn_it = functions.begin();
	while ( fn_it != fn_it_end ) {
		fn_it->SavePLA( "to_min.pla" );
		fn_it = functions.erase(fn_it);
		system( "espresso to_min.pla > from_min.pla" );
		pla.LoadPLA( "from_min.pla" );
		functions.push_front(pla);
	}
}

void SaveGenlib( list<PLA> &functions, char *fn )
{
	unsigned int i;
	ofstream f;
	list<PLA>::iterator fn_it;
	list<Term>::const_iterator ti;
	unsigned int cnt = 1;
	char c;
	bool flag_t, flag_l;

	f.open(fn);
	f << "# generated by GenlibGen (c) by Petr Fiser" << endl;
	for ( fn_it = functions.begin(); fn_it != functions.end(); ++fn_it ) {
		if ( !fn_it->isConstant() ) {
			f << "GATE     cell" << cnt++ << "     1.00  O=";
			flag_t = false;
			for ( ti = fn_it->getConstDataRef().begin(); ti != fn_it->getConstDataRef().end(); ++ti ) {
				if ( ti->isOM_1(0) ) {
					flag_l = false;
					if ( flag_t ) f << "+";
					for ( i = 0; i < fn_it->inputs(); i++ ) {
						if ( !ti->isLit_dc(i) ) {
							if ( flag_l ) f << "*";
							if ( ti->isLit_0(i) ) f << "!";
							c = 'a' + i;
							f << c;
							flag_t = true;
							flag_l = true;
						}
					}
				}
			}
			f << ";" << endl;
			f << "   PIN   *  UNKNOWN  1  999  1.00  0.00  1.00  1.00" << endl;
		}
	}
	f << "GATE     one         0.00  O=CONST1;" << endl;
	f << "GATE     zero        0.00  O=CONST0;" << endl;
	f.close();
}

int main(int argc, char* argv[])
{
	int i;
	unsigned int j;
	unsigned int inputs = 0;
	unsigned int classes = 0;
	char *oname = NULL;
	list<PLA> functions;
	
	cerr << "GenlibGen 1.0 (c) Petr Fiser, compiled " << __DATE__ << ", " << __TIME__ << endl;
    if ( argc < 3 ) {
		Help();
		exit(1);
    }

	i = 1;
	while ( i < argc ) {
		if ( strcmp( argv[i], "-all" ) == 0 ) {
			classes = 0;
			i++;
		}
		else if ( strcmp( argv[i], "-P" ) == 0 ) {
			classes = 1;
			i++;
		}
		else if ( strcmp( argv[i], "-NP" ) == 0 ) {
			classes = 2;
			i++;
		}
		else if ( strcmp( argv[i], "-NPN" ) == 0 ) {
			classes = 3;
			i++;
		}
		else if ( strcmp( argv[i], "-i" ) == 0 ) {
			i++;
			inputs = atoi(argv[i]);
			i++;
		}
		else if ( argv[i][0] != '-' && oname == NULL ) {
			oname = argv[i];
			i++;
		}
		else {
			Help();
			exit(1);
		}
    }

	if ( inputs == 0 || oname == NULL ) {
		Help();
		exit(1);
    }

	GenerateAll(functions, inputs);
	Minimize(functions);
	SaveGenlib(functions, oname);

    return 0;
}
