//---------------------------------------------------------------------------

#ifndef CubeH
#define CubeH

//#include "cubegraph.h"

typedef enum {ZERO, ONE, DC, UNDEF} Literal;
#include <unordered_set>


class CubeStd {
  protected:
	int dim;            // dimension. -1 means that it is not computed yet. Just for internal purposes.
    string IM;           // the term. '0', '1' and '-' are valid values. Deprecated
	pair<bitvector, bitvector> IM_data;		// (00) = 0, (10) = 1, (01) = dc

    unsigned int index;  // index of the term
	bool unwanted;		// this one is to be deleted by Purge();

		// for BOOM
	vector<bitvector> SO_CoverVector;	// vector of source function terms, which are covered by it, for each output separately
	bitvector CoverVector;	// vector of source function terms, which are covered by it

  public:

    CubeStd(void);
    CubeStd( const unsigned int iv );
    ~CubeStd(void);
	void CopyFrom(const CubeStd* src);


	bool isEmpty() const;

	int tmp;					// for public use. Must not be modified by the kernel!

	bool operator== ( const string &s ) const { return this->IM == s; };
	bool operator== ( const CubeStd &t ) const;

	void clear(void);				// clears all data
	void init( const unsigned int iv );	// assigns all dc values
		
	CubeStd getCube3() const;

		// these should be used with care! For PLA and SoPLA only.	
	void setIndex( const unsigned int val ) { index = val; };
	unsigned int getIndex(void) const { return index; };		// !Needs not be correctly computed! Do use PLA.getIndex() instead

		// setters and getters
	void setLit_0( const unsigned int pos );				// sets one bit in IM to 1
	void setLit_1( const unsigned int pos );				// sets one bit in IM to 0
	void setLit_dc( const unsigned int pos );				// sets one bit in IM to dc
	void setLit( const unsigned int pos, const Literal val );
	void setLit( const unsigned int pos, const char val );      // sets one value in the IM
	bool isLit_0( const unsigned int pos ) const;			// is the given IM bit 0?
	bool isLit_1( const unsigned int pos ) const;			// is the given IM bit 1?
	bool isLit_dc( const unsigned int pos ) const;			// is the given IM bit dc?
	Literal getLit2( const unsigned int pos ) const;	// new version. Will replace getLit
	vector<Literal> getCube2(void) const;				// new version. Will replace getCube
	void assign( Literal val );						// assigns a value to all vector items
	void assign( unsigned int size, Literal val );


	// deprecated
    string getCube(void) const { return IM; };
    void setCube( const string val );      // sets IM
    void setCube( const vector<Literal> val );      // sets IM
    char getLit( const unsigned int pos ) const { return IM[pos]; };

    unsigned int getDim(void);
	void decDim(int cnt) { if ( dim != -1 ) dim -= cnt; }
	unsigned int getLiterals(void) { return IM.size()-getDim(); };
	void setUnwanted(void) { unwanted = true; };
	bool isUnwanted(void) const { return unwanted; };
	unsigned int size(void) const { return IM_data.first.size(); };		// returns # of term variables
	unsigned int inputs(void) const { return IM_data.first.size(); };		// returns # of term variables

	void getCoverVector( const bitvector _CoverVector ) { CoverVector = _CoverVector; };
	bitvector *getCoverVector(void) { return &CoverVector; };
	void setSO_CoverVector( const vector<bitvector> _SO_CoverVector ) { SO_CoverVector = _SO_CoverVector; };
	bitvector *getSO_CoverVector( const unsigned int ov ) { return &SO_CoverVector[ov]; };
	vector<bitvector> *getSO_CoverVector() { return &SO_CoverVector; };

    bool intersects( const string &s ) const;				// does it intersect a cube given by a string? 
	bool intersects( const CubeStd &t ) const;				// does it intersect a cube? 
    bool contains( const string &s ) const;		// does it contain s?
	bool contains( const CubeStd &t ) const;		// does it contain t?
	CubeStd intersection( const string &s ) const;				// returns an intersection of two terms. The intersection must exist
	CubeStd intersection(  const CubeStd &t ) const;				// returns an intersection of two terms. The intersection must exist
	int intersectSize( const string &s ) const;
	int intersectSize( const CubeStd &t ) const;

	bool And( const string &s ) const;
	bool And( CubeStd &t );

};


class CubeGraph {
public:
	typedef std::tr1::unordered_set<int> Container;
	//int dim;            // dimension. -1 means that it is not computed yet. Just for internal purposes.

  private:
    //string IM;           // the term. '0', '1' and '-' are valid values. Deprecated
	//pair<bitvector, bitvector> IM_data;		// (00) = 0, (10) = 1, (01) = dc
	Container one, zero;
	int litCount;

    unsigned int index;  // index of the term
	bool unwanted;		// this one is to be deleted by Purge();

		// for BOOM
	vector<bitvector> SO_CoverVector;	// vector of source function terms, which are covered by it, for each output separately
	bitvector CoverVector;	// vector of source function terms, which are covered by it

  public:
	  bool isUnwanted() {return unwanted;}

    CubeGraph(void);
    CubeGraph( const unsigned int iv );
	CubeGraph( const CubeGraph &src);
	void operator= ( const CubeGraph &src );
    ~CubeGraph(void);
	void CopyFrom(const CubeGraph* src);

	bool isEmpty() const;

	int tmp;					// for public use. Must not be modified by the kernel!

	bool operator== ( const string &s ) const;
	bool operator== ( const CubeGraph &t ) const;

	void clear(void);				// clears all data
	void init( const unsigned int iv );	// assigns all dc values

		// these should be used with care! For PLA and SoPLA only.	
	void setIndex( const unsigned int val ) { index = val; };
	unsigned int getIndex(void) const { return index; };		// !Needs not be correctly computed! Do use PLA.getIndex() instead

		// setters and getters
	inline void setLit_0( const unsigned int pos );				// sets one bit in IM to 1
	inline void setLit_1( const unsigned int pos );				// sets one bit in IM to 0
	inline void setLit_dc( const unsigned int pos );			// sets one bit in IM to dc
	void setLit( const unsigned int pos, const Literal val );
	void setLit( const unsigned int pos, const char val );      // sets one value in the IM
	inline bool isLit_0( const unsigned int pos ) const;			// is the given IM bit 0?
	inline bool isLit_1( const unsigned int pos ) const;			// is the given IM bit 1?
	inline bool isLit_dc( const unsigned int pos ) const;			// is the given IM bit dc?
	Literal getLit2( const unsigned int pos ) const;	// new version. Will replace getLit
	vector<Literal> getCube2(void) const;				// new version. Will replace getCube
	void assign( Literal val );						// assigns a value to all vector items
	void assign( unsigned int size, Literal val );


	// deprecated
    string getCube(void) const;// { return IM; };
	CubeGraph getCube3() const;
    void setCube( const string val );      // sets IM
    void setCube( const vector<Literal> val );      // sets IM
    char getLit( const unsigned int pos ) const;// { return IM[pos]; };

    unsigned int getDim(void);
	void decDim(int cnt) {}
	unsigned int getLiterals(void) { return size()-getDim(); };
	void setUnwanted(void) { unwanted = true; };
	bool isUnwanted(void) const { return unwanted; };
	unsigned int size(void) const;		// returns # of term variables
	unsigned int inputs(void) const;		// returns # of term variables

	void getCoverVector( const bitvector _CoverVector ) { CoverVector = _CoverVector; };
	bitvector *getCoverVector(void) { return &CoverVector; };
	void setSO_CoverVector( const vector<bitvector> _SO_CoverVector ) { SO_CoverVector = _SO_CoverVector; };
	bitvector *getSO_CoverVector( const unsigned int ov ) { return &SO_CoverVector[ov]; };
	vector<bitvector> *getSO_CoverVector() { return &SO_CoverVector; };

    bool intersects( const string &s ) const;				// does it intersect a cube given by a string? 
	bool intersects( const CubeGraph &t ) const;				// does it intersect a cube? 
    bool contains( const string &s ) const;		// does it contain s?
	bool contains( const CubeGraph &t ) const;		// does it contain t?
	CubeGraph intersection( const string &s ) const;				// returns an intersection of two terms. The intersection must exist
	CubeGraph intersection(  const CubeGraph &t ) const;				// returns an intersection of two terms. The intersection must exist
	int intersectSizeStr( const CubeGraph &t ) const;
	int intersectSize( const CubeGraph &t ) const;

	bool And( const string &s ) const;
	bool And( CubeGraph &t );

	friend class Cube;
	friend class FullGraph;
};



typedef CubeGraph CubeBase;

class Cube : public CubeBase {
public:
	Cube(void) : CubeBase() {}
	Cube( const unsigned int iv ) : CubeBase(iv) {}
	Cube getCube3() const;

};

typedef list<Cube> CubeList;

//---------------------------------------------------------------------------
#endif
