//---------------------------------------------------------------------------

#include "stdafx.h"

using namespace std;

#ifdef _MSC_VER
#pragma hdrstop
#endif

#include "myExcept.h"
#include "bitvector.h"
#include "Cube.h"
#include "mytime.h"

#include <ctime>
#include <cstdio>
#include "perfcounter.h"

//// MACROS
///////////
//#define FULLGRAPH

#ifdef FULLGRAPH
    #define FULL_SET_COUNT(count) g_fullGraph.setVariableCount(count)
    #define FULL_SET_TERM(variable, cube) g_fullGraph.setTerm(variable, cube)
    #define FULL_REMOVE_TERM(variable, cube) g_fullGraph.removeTerm(variable, cube)
    #define FULL_SET_ALL(cube) g_fullGraph.setAllVars(cube)
    #define FULL_REMOVE_ALL(cube) g_fullGraph.removeAllVars(cube)
#else
    #define FULL_SET_COUNT(count)
    #define FULL_SET_TERM(variable, cube)
    #define FULL_REMOVE_TERM(variable, cube)
    #define FULL_SET_ALL(cube)
    #define FULL_REMOVE_ALL(cube)
#endif


//// FULL GRAPH CLASS
/////////////////////
class FullGraph {
public:
    typedef std::tr1::unordered_set<CubeGraph*> VarTerms;
    struct Sets {
        VarTerms one;
        VarTerms zero;
        VarTerms dc;
    };
    // all terms with given variable
    typedef vector<Sets> Variables;

    void setVariableCount(int count) {
        vars.resize(count);
    }

    void setTerm(int variable, CubeGraph* cube) {
        removeTerm(variable, cube);
        switch (cube->getLit(variable)) {
            case '1': vars[variable].one.insert(cube); break;
            case '0': vars[variable].zero.insert(cube); break;
            case '-': vars[variable].dc.insert(cube); break;
        };
    }
    void removeTerm(int variable, CubeGraph* cube) {
        vars[variable].one.erase(cube);
        vars[variable].zero.erase(cube);
        vars[variable].dc.erase(cube);
    }
    void setAllVars(CubeGraph* cube) {
        for (int i = 0; i<cube->size(); i++)
            switch (cube->getLit(i)) {
                case '1': vars[i].one.insert(cube); break;
                case '0': vars[i].zero.insert(cube); break;
                case '-': vars[i].dc.insert(cube); break;
            };
    }
    void removeAllVars(CubeGraph* cube) {
        for (int i = 0; i<vars.size(); i++) {
            vars[i].one.erase(cube);
            vars[i].zero.erase(cube);
            vars[i].dc.erase(cube);
        }
    }

protected:
    Variables vars;

} g_fullGraph;



//// INITIALIZATION METHODS
///////////////////////////

CubeGraph::CubeGraph(void)
{
    unwanted = false;
    litCount = 0;
}

CubeGraph::CubeGraph( const unsigned int iv )
{
    init(iv);
}

void CubeGraph::operator= ( const CubeGraph &src ) {
    FULL_REMOVE_ALL(this);
    one = src.one;
    zero = src.zero;
    litCount = src.litCount;
    index = src.index;
    unwanted = src.unwanted;
    SO_CoverVector = src.SO_CoverVector;
    CoverVector = src.CoverVector;
    tmp = src.tmp;
    FULL_SET_ALL(this);
}

CubeGraph::CubeGraph( const CubeGraph &src) {
    operator=(src);
}

CubeGraph::~CubeGraph(void)
{
    FULL_SET_ALL(this);
}

void CubeGraph::clear(void)
// Clears the CubeGraph structure
{
    SO_CoverVector.clear();
    CoverVector.clear();
    one.clear();
    zero.clear();
}

void CubeGraph::init( const unsigned int iv )
{
    FULL_SET_COUNT(iv);
    FULL_REMOVE_ALL(this);
    one.clear();
    zero.clear();
    unwanted = false;
    SO_CoverVector.clear();
    CoverVector.clear();
    litCount = iv;
}


void CubeGraph::setCube( const string val )
{
    FULL_SET_COUNT(val.size());
    FULL_REMOVE_ALL(this);
     unsigned int i;

    if ( val.size() != this->size() && this->size() != 0 ) throw RuntimeException( "CubeGraph::setCube: Wrong IM size" );
    one.clear();
    zero.clear();
    litCount = val.size();
    for ( i = 0; i < val.size(); i++ )
        if (val[i] != '-')
            this->setLit( i, val[i] );    
}

void CubeGraph::setCube( const vector<Literal> val )
{
    FULL_SET_COUNT(val.size());
    FULL_REMOVE_ALL(this);
     unsigned int i;

    if ( val.size() != this->size() && this->size() != 0 ) throw RuntimeException( "CubeGraph::setCube: Wrong IM size" );
    litCount = val.size();
    one.clear();
    zero.clear();
    for ( i = 0; i < val.size(); i++ )
        if (val[i] != DC)
            this->setLit( i, val[i] );    
}

void CubeGraph::assign( Literal val )
// assigns a value to all vector items
{
    FULL_REMOVE_ALL(this);
     unsigned int n;

    one.clear();
    zero.clear();

    n = litCount;
    switch ( val ) {
        case ONE: for (int i = 0; i<litCount ; i++) one.insert(i);
                  FULL_SET_ALL(this);
                  break;
        case ZERO:for (int i = 0; i<litCount ; i++) zero.insert(i);
                  FULL_SET_ALL(this);
                  break;
        case DC:  break;
        default: throw RuntimeException( "CubeGraph::assign: incorrect literal value" );
    }
}

void CubeGraph::assign( unsigned int size, Literal val )
// assigns a value to all vector items
{
    litCount = size;
    this->assign(val);
}

void CubeGraph::CopyFrom(const CubeGraph* src) {
    FULL_REMOVE_ALL(this);
    one = src->one;
    zero = src->zero;
    litCount = src->litCount;
    FULL_SET_ALL(this);
}



//// GET/GET METHODS
////////////////////

string CubeGraph::getCube(void) const {
    string s(size(), '-');
    for (Container::const_iterator it = one.begin();
        it != one.end(); ++it)
            s[*it] = '1';
    for (Container::const_iterator it = zero.begin();
        it != zero.end(); ++it)
            s[*it] = '0';
    return s;
}

vector<Literal> CubeGraph::getCube2(void) const
{
    vector<Literal> res;
    unsigned int i;

    for ( i = 0; i < size(); i++ )
        res.push_back( getLit2(i) );
    return res;
}

CubeGraph CubeGraph::getCube3() const {
    CubeGraph c;
    c.CopyFrom(this);
    return c;
}

bool CubeGraph::isEmpty() const {
    return one.empty() && zero.empty();
}

unsigned int CubeGraph::size(void) const { return litCount; }
unsigned int CubeGraph::inputs(void) const { return litCount; }

void CubeGraph::setLit_0( const unsigned int pos )
{
    if ( pos >= this->size() ) throw RuntimeException( "CubeGraph::setLit_0: Index out of range" );
    FULL_SET_TERM(pos, this);
    one.erase(pos);
    zero.insert(pos);
}

void CubeGraph::setLit_1( const unsigned int pos )
{
    if ( pos >= this->size() ) throw RuntimeException( "CubeGraph::setLit_1: Index out of range" );
    FULL_SET_TERM(pos, this);
    one.insert(pos);
    zero.erase(pos);
}

void CubeGraph::setLit_dc( const unsigned int pos )
{
    if ( pos >= this->size() ) throw RuntimeException( "CubeGraph::setLit_dc: Index out of range" );
    FULL_SET_TERM(pos, this);
    one.erase(pos);
    zero.erase(pos);
}

bool CubeGraph::isLit_0( const unsigned int pos ) const
{
    return zero.find(pos) != zero.end();
}

bool CubeGraph::isLit_1( const unsigned int pos ) const
{
    return one.find(pos) != one.end();
}

bool CubeGraph::isLit_dc( const unsigned int pos ) const
{
    return (!isLit_0(pos)) && (!isLit_1(pos));
}

void CubeGraph::setLit( const unsigned int pos, const Literal val  )
{
     if ( pos >= this->size() ) throw RuntimeException( "CubeGraph::setLit: Index out of range" );
    switch ( val ) {
        case ZERO:
            setLit_0(pos);
            break;
        case ONE:
            setLit_1(pos);
            break;
        case DC:
            setLit_dc(pos);
            break;
        default: throw RuntimeException( "CubeGraph::setLit: Wrong IM value" );
            break;
    }
}

void CubeGraph::setLit( const unsigned int pos, const char val )
// Sets a given input
{
     if ( pos >= this->size() ) throw RuntimeException( "CubeGraph::setLit: Index out of range" );
    switch ( val ) {
        case '0':
            this->setLit_0(pos);
            break;
        case '1':
            this->setLit_1(pos);
            break;
        case '-':
            this->setLit_dc(pos);
            break;
        default:
            throw RuntimeException( "CubeGraph::setLit: Invalid value" );
    }
}


char CubeGraph::getLit( const unsigned int pos ) const {
    if (isLit_0(pos))
        return '0';
    else if (isLit_1(pos))
        return '1';
    else return '-';
};

Literal CubeGraph::getLit2( const unsigned int pos ) const
{
     if ( this->isLit_0(pos) ) return ZERO;
    else if ( this->isLit_1(pos) ) return ONE;
    else return DC;
}

unsigned int CubeGraph::getDim(void)
// Returns term dimension. If it is not already computed, computes it
{
     return litCount - one.size() - zero.size();
}



//// CUBE OPERATIONS
////////////////////

bool CubeGraph::intersects( const CubeGraph &t ) const        // this is much faster
/*
    Returns: do the terms intersect?
*/
{
MESURE
    unsigned int i;

    if ( t.size() != this->size() ) throw RuntimeException("CubeGraph::intersects: Comparing incompatible terms");

    bool thisSmaller = (this->one.size() + this->zero.size()) <
                        (t.one.size() + t.zero.size());
    const CubeGraph& first = thisSmaller ? *this : t;
    const CubeGraph& second = thisSmaller ? t : *this;

    for (Container::const_iterator it = first.one.begin();
        it != first.one.end(); ++it)
    {
        if (second.zero.find(*it) != second.zero.end())
            return false;
    }

    for (Container::const_iterator it = first.zero.begin();
        it != first.zero.end(); ++it)
    {
        if (second.one.find(*it) != second.one.end())
            return false;
    }

    return true;
}

int CubeGraph::intersectSize( const CubeGraph &t ) const
{
    throw RuntimeException("this method is not tested, may return invalid result");

    int ret = 0;

    for (Container::const_iterator it = one.begin(); it != one.end(); ++it)
        if (t.isLit_dc(*it))
            ret++;
    for (Container::const_iterator it = zero.begin(); it != zero.end(); ++it)
        if (t.isLit_dc(*it))
            ret++;
    for (Container::const_iterator it = t.one.begin(); it != t.one.end(); ++it)
        if (this->isLit_dc(*it))
            ret++;
    for (Container::const_iterator it = t.zero.begin(); it != t.zero.end(); ++it)
        if (this->isLit_dc(*it))
            ret++;

    return ret;
}


bool CubeGraph::contains( const CubeGraph &t ) const        // this is surprisingly faster
//  Returns: does it contain t?
{
MESURE
    unsigned int i;

    if ( t.size() != this->size() )
        throw RuntimeException("CubeGraph::contains: Comparing incompatible terms");

    for (Container::const_iterator it = one.begin();
        it != one.end(); ++it)
    {
        if (t.one.find(*it) == t.one.end())
            return false;
    }

    for (Container::const_iterator it = zero.begin();
        it != zero.end(); ++it)
    {
        if (t.zero.find(*it) == t.zero.end())
            return false;
    }
    return true;
}

bool CubeGraph::operator== ( const string &s ) const {
    return getCube() == s;
};


bool CubeGraph::operator== ( const CubeGraph &t ) const
// returns true, if the cubes are equal
{
    if (litCount != t.litCount || one.size() != t.one.size() ||
        zero.size() != t.zero.size())
            return false;
    
    for (Container::const_iterator it = one.begin();
        it != one.end(); ++it)
            if (t.one.find(*it) == t.one.end())
                return false;
    for (Container::const_iterator it = zero.begin();
        it != zero.end(); ++it)
            if (t.zero.find(*it) == t.zero.end())
                return false;
    return true;
}



/*
//// TESTING METHODS
////////////////////

bool CubeGraph::And( const string &s ) const
{
    unsigned int i;
    bool res = true;

    if ( s.size() != this->IM.size() )
        throw RuntimeException("CubeGraph::contains: Comparing incompatible terms");

    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] == '1' && s[i] != '1' ) res = false;
        // davidtodo - rovnou return false;
    return res;
}


bool CubeGraph::And( CubeGraph &t )
{
    //return ( this->IM_data.first & t.IM_data.first ).isOne();
    t.IM_data.first &= this->IM_data.first;
    return t.IM_data.first.isOne();
}


bool CubeGraph::contains( const string &s ) const
{
    unsigned int i;

    if ( s.size() != this->IM.size() )
        throw RuntimeException("CubeGraph::contains: Comparing incompatible terms");

    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] != s[i] && this->IM[i] != '-' ) return false;
    return true;
}


CubeGraph CubeGraph::intersection( const CubeGraph &t ) const
{
    // pure graph
    CubeGraph inter;
    inter.one = this->one;
    for (Container::const_iterator it = t.one.begin();
        it != t.one.end(); ++it)
            inter.one.insert(*it);

    inter.zero = this->zero;
    for (Container::const_iterator it = t.zero.begin();
        it != t.zero.end(); ++it)
            inter.zero.insert(*it);
    return inter;

}

CubeGraph CubeGraph::intersection( const string& s ) const
{
    unsigned int i;
    CubeGraph t1;

    t1 = *this;
    for ( i = 0; i < IM.size(); i++ ) {
        if ( t1.IM[i] != s[i] && t1.IM[i] != '-' && s[i] != '-' ) throw RuntimeException( "CubeGraph::intersection: no intersection" );;
        if ( s[i] == '0' || s[i] == '1' ) t1.IM[i] = s[i];
    }
    return t1;
}

int CubeGraph::intersectSizeStr( const CubeGraph &t ) const
{
    const string &s = t.IM;
    unsigned int i;
    int j = 0;

    if ( s.size() != this->IM.size() ) throw RuntimeException("CubeGraph::intersects: Comparing incompatible terms");
    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] != s[i] && this->IM[i] != '-' && s[i] != '-' ) return -1;
        else if ( (this->IM[i] == '-' && s[i] != '-') || (this->IM[i] != '-' && s[i] == '-') ) j++;
    return j;
}

bool CubeGraph::intersects( const string &s ) const
{
    unsigned int i;

    if ( s.size() != this->size() ) throw RuntimeException("CubeGraph::intersects: Comparing incompatible terms");
    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] != s[i] && this->IM[i] != '-' && s[i] != '-' ) return false;
    return true;
}
*/

