#include "stdafx.h"

#ifdef _MSC_VER
#pragma hdrstop
#endif

using namespace std;

#include "myExcept.h"
#include "bitvector.h"


bitvector::bitvector(void)
{
	bitsize = 0;
}

bitvector::bitvector( const unsigned int _size )
{
	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	data.resize(items(), 0);
}

bitvector::~bitvector(void)
{
}

void bitvector::resize( const unsigned int _size )
{
	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	data.resize(items(), 0);
}

void bitvector::resize( const unsigned int _size, const bool val )
// resizes the vector and assignes a value to new vector items
{
	DATATYPE d_val;

	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	d_val = (val)?ALL_ONE:0;
	data.resize(items(), d_val);
}

void bitvector::assign( const unsigned int _size, const bool val )
// assignes a value to all vector items
{
	DATATYPE d_val;

	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	d_val = (val)?ALL_ONE:0;
	data.assign(items(), d_val);
}

position bitvector::GetPosition( const unsigned int pos ) const
{
	position p;

	p.index = (pos >> SHIFT);
	p.bit = pos & MASK;

	return p;
}

void bitvector::push_back( const bool bit )
{
	position p;

	p = GetPosition( bitsize );
	if ( p.bit == 0 ) data.resize(data.size()+1, 0);
	bitsize++;
	if ( bit ) data[p.index] |= (1 << p.bit);
	else data[p.index] &= ~(1 << p.bit);
}

bool bitvector::operator[] ( const unsigned int pos ) const
{
	position p;

	if ( pos >= bitsize ) throw RuntimeException( "bitvector:: index out of range" );
	p = GetPosition(pos);
	return( (data[p.index] & (1 << p.bit)) != 0 );
}

bool bitvector::operator== ( const bitvector &_vector ) const
{
	unsigned int i;
	position p;
	DATATYPE mask;

	if ( this->bitsize != _vector.bitsize ) return false;
	else {
		for ( i = 0; i < this->items()-1; i++ )
			if ( this->data[i] != _vector.data[i] ) return false;
		p = GetPosition(bitsize-1);
		mask = ~(ALL_ONE << (p.bit+1));
		if ( (this->data[i] & mask) != (_vector.data[i] & mask) ) return false;
	}
	return true;
}

void bitvector::clear(void)
{
	data.clear();
	bitsize = 0;
}

void bitvector::setbit( const unsigned int pos, const bool val )
{
	position p;

	if ( pos > bitsize-1 ) throw RuntimeException( "bitvector:: index out of range" );
	p = GetPosition( pos );
	if ( !val ) data[p.index] &= ~(1 << p.bit);
	else data[p.index] |= (1 << p.bit);
}

bitvector bitvector::operator& ( const bitvector &_vector ) const
{
	bitvector vector;
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = this->data[i] & _vector.data[i];
	return vector;
}

bitvector bitvector::operator| ( const bitvector &_vector ) const
{
	bitvector vector;
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = this->data[i] | _vector.data[i];
	return vector;
}

bitvector bitvector::operator^ ( const bitvector &_vector ) const
{
	bitvector vector;
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = this->data[i] ^ _vector.data[i];
	return vector;
}

bitvector bitvector::operator~(void) const
{
	bitvector vector;
	unsigned int i;

	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = ~this->data[i];
	return vector;
}

bool bitvector::contains( const bitvector &_vector ) const
// returns true, if this contains _vector
// i.e., _vector does not have a true value, where this has
{
	unsigned int i;
	position p;
	DATATYPE mask;

	if ( this->size() != _vector.size() )
			throw RuntimeException( "bitvector::contains: comparing vectors of different sizes" );
	for ( i = 0; i < this->items()-1; i++ )
		if ( (this->data[i] | _vector.data[i]) != this->data[i] ) return false;
	p = GetPosition(bitsize-1);
	mask = ~(ALL_ONE << (p.bit+1));
	if ( ((this->data[i] & mask) | _vector.data[i]) != (this->data[i] & mask) ) return false;
	return true;
}

bool bitvector::isNull(void) const
{
	unsigned int i;
	position p;
	DATATYPE mask;

	for ( i = 0; i < this->items()-1; i++ )
		if ( data[i] != 0 ) return false;
	p = GetPosition(bitsize-1);
	mask = ~(ALL_ONE << (p.bit+1));
	if ( (this->data[i] & mask) != 0 ) return false;
	return true;
}

bool bitvector::isOne(void) const
{
	unsigned int i;
	position p;
	DATATYPE mask;

	for ( i = 0; i < this->items()-1; i++ )
		if ( data[i] != ALL_ONE ) return false;
	p = GetPosition(bitsize-1);
	mask = ~(ALL_ONE << (p.bit+1));
	if ( (this->data[i] & mask) != mask ) return false;
	return true;
}

void bitvector::operator&= ( const bitvector &_vector )
{
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	for ( i = 0; i < this->items(); i++ ) this->data[i] &= _vector.data[i];
}

void bitvector::operator|= ( const bitvector &_vector )
{
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	for ( i = 0; i < this->items(); i++ ) this->data[i] |= _vector.data[i];
}

void bitvector::operator^= ( const bitvector &_vector )
{
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	for ( i = 0; i < this->items(); i++ ) this->data[i] ^= _vector.data[i];
}
