//---------------------------------------------------------------------------

#include "stdafx.h"

using namespace std;

#pragma hdrstop

#include "myExcept.h"
#include "bitvector.h"
#include "Term.h"

//---------------------------------------------------------------------------


Term::Term(void)
{
    dim = -1;
	unwanted = false;
	origin = 0;
}

Term::Term( const unsigned int iv, const unsigned int ov )
{
    unsigned int i;

    IM.resize( iv );
    OM.resize( ov );
    for ( i = 0; i < iv; i++ ) IM[i] = '-';
    for ( i = 0; i < ov; i++ ) OM[i] = '~';
    dim = iv;
	unwanted = false;
}

Term::~Term(void)
{
}

void Term::SetIM( const unsigned int pos, const char val )
{
	if ( pos >= IM.size() ) throw RuntimeException( "Term::SetIM: Index out of range" );
    if ( val != '0' && val != '1' && val != '-' ) throw RuntimeException( "Term::SetIM: Invalid value" );
    if ( dim != -1 && IM[pos] == '-' && val != '-' ) dim--;
    if ( dim != -1 && IM[pos] != '-' && val == '-' ) dim++;
    IM[pos] = val;
}

void Term::SetOM( const unsigned int pos, const char val )
{
    if ( pos >= OM.size() ) throw RuntimeException( " Term::SetOM: Index out of range" );
    if ( val != '0' && val != '1' && val != '-' && val != '~' ) throw RuntimeException( "Term::SetOM: Invalid value" );
    OM[pos] = val;
}

void Term::SetIM( const string val )
{
	if ( val.size() != IM.size() && IM.size() != 0 ) throw RuntimeException( " Term::SetIM: Wrong IM size" );
	IM = val;
	dim = -1;
}

void Term::SetOM( const string val )
{
	if ( val.size() != OM.size() && OM.size() != 0 ) throw RuntimeException( " Term::SetOM: Wrong IM size" );
	OM = val;
}

unsigned int Term::GetDim(void)
// Returns dimension. If it is not already computed, computes it

{
    unsigned int i, j;

    if ( dim != -1 ) return dim;
    i = 0;
    for ( j = 0; j < IM.size(); j++ )
        if ( IM[j] == '-' ) i++;
    dim = i;
    return i;
}

/*
void Term::DiscardDim()
{
    dim = -1;
}*/

unsigned int Term::GetOC(void) const
// Returns the output cost.

{
    unsigned int j, oc;
    oc = 0;
    for ( j = 0; j < OM.size(); j++ )
        if ( OM[j] == '1' ) oc++;
    return oc;
}

bool Term::Intersects( const string& s ) const
/*
    Returns: do the terms intersect?
*/
{
    unsigned int i;

	if ( s.size() != this->IM.size() ) throw RuntimeException("Term::Intersects: Comparing incompatible terms");
    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] != s[i] && this->IM[i] != '-' && s[i] != '-' ) return false;
    return true;
}

int Term::IntersectSize( const string& s ) const
/*
    Returns: do the terms intersect?
*/
{
    unsigned int i;
	int j = 0;

	if ( s.size() != this->IM.size() ) throw RuntimeException("Term::Intersects: Comparing incompatible terms");
    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] != s[i] && this->IM[i] != '-' && s[i] != '-' ) return -1;
		else if ( (this->IM[i] == '-' && s[i] != '-') || (this->IM[i] != '-' && s[i] == '-') ) j++;
    return j;
}

bool Term::Intersects( const Term& t ) const
/*
    Returns: do the terms intersect?
    Checks also outputs
*/
{
    unsigned int i;

    if ( t.IM.size() != this->IM.size() || t.OM.size() != this->OM.size() )
     throw RuntimeException("Term:Intersects:Comparing incompatible terms");
           // do the outputs intersect?
    for ( i = 0; i < OM.size(); i++ )
        if ( this->OM[i] == '1' && t.OM[i] == '1' ) break;
    if ( i == OM.size() ) return false;
           // do the intputs intersect?
    return ( Intersects(t.IM) );
}

int Term::IntersectSize( const Term& t ) const
/*
    Returns: do the terms intersect?
    Checks also outputs
*/
{
    unsigned int i;

    if ( t.IM.size() != this->IM.size() || t.OM.size() != this->OM.size() )
		throw RuntimeException("Term::Intersects: Comparing incompatible terms");
           // do the outputs intersect?
    for ( i = 0; i < OM.size(); i++ )
        if ( this->OM[i] == '1' && t.OM[i] == '1' ) break;
    if ( i == OM.size() ) return -1;
           // do the intputs intersect?
    return ( IntersectSize(t.IM) );
}

unsigned int Term::Relation( const Term& t ) const
/*
    Bit 0: non-empty IM intersection
	Bit 1: t.IM is contained
	Bit 2: IMs are equal
    Bit 3: non-empty OM intersection
	Bit 4: t.OM is contained
	Bit 5: OMs are equal
*/
{
    unsigned int i;
	int res = 63-8;

	if ( t.IM.size() != this->IM.size() || t.OM.size() != this->OM.size() )
     throw RuntimeException("Term:Intersects:Comparing incompatible terms");

	for ( i = 0; i < OM.size(); i++ ) {
        if ( this->OM[i] != '1' && t.OM[i] == '1' ) res &= (63-16);   // t is not contained
		if ( this->OM[i] == '1' && t.OM[i] == '1' ) res |= 8;		// non-empty intersection
		if ( this->OM[i] != t.OM[i] ) res &= (63-32);		// not equal
	}
	for ( i = 0; i < IM.size(); i++ ) {
		if ( abs(this->IM[i] - t.IM[i]) == 1 ) {
			res &= (63-7);   // 0/1 conflict. No intersection
			break;
		}
		if ( this->IM[i] != '-' && t.IM[i] == '-' ) res &= (63-2);		// not contained
		if ( this->IM[i] != t.IM[i] ) res &= (63-4);		// not equal
	}
    return ( res );
}

unsigned int Term::Relation2( const Term& t ) const
/*
    0: non-empty IM intersection, terminates
	Bit 1: t.IM is contained
	Bit 2: IMs are equal
    Bit 3: non-empty OM intersection
	Bit 4: t.OM is contained
	Bit 5: OMs are equal
*/
{
    unsigned int i;
	int res = 63-8;

	if ( t.IM.size() != this->IM.size() || t.OM.size() != this->OM.size() )
       throw RuntimeException("Term:Intersects:Comparing incompatible terms");

	for ( i = 0; i < IM.size(); i++ ) {
		if ( abs(this->IM[i] - t.IM[i]) == 1 ) {
			res = 0;   // 0/1 conflict. No intersection. Terminate
			break;
		}
		if ( this->IM[i] != '-' && t.IM[i] == '-' ) res &= (63-2);		// not contained
		if ( this->IM[i] != t.IM[i] ) res &= (63-4);		// not equal
	}
	if (res != 0) 
	  for ( i = 0; i < OM.size(); i++ ) {
        if ( this->OM[i] != '1' && t.OM[i] == '1' ) res &= (63-16);   // t is not contained
		if ( this->OM[i] == '1' && t.OM[i] == '1' ) res |= 8;		// non-empty intersection
		if ( this->OM[i] != t.OM[i] ) res &= (63-32);		// not equal
	  }
    return res;
}

bool Term::Intersects( const Term& t, const unsigned int _ov ) const
/*
    Returns: do the terms intersect?
    Checks given output only
*/
{
    if ( t.IM.size() != this->IM.size() || t.OM.size() != this->OM.size() )
     throw RuntimeException("Term:Intersects:Comparing incompatible terms");
           // do the outputs intersect?
    if ( this->OM[_ov] == '1' && this->OM[_ov] != t.OM[_ov] ) return false;
           // do the intputs intersect?
    return ( Intersects(t.IM) );
}



bool Term::Contains_noOM_Check( const Term& t ) const
/*
    Returns: does it contain t? OM size not checked
    Only for f-type PLA
*/
{
    unsigned int i;

    if ( t.IM.size() != this->IM.size() )
		throw RuntimeException("Term::Contains: Comparing incompatible terms");

    for ( i = 0; i < IM.size(); i++ )
        if ( this->IM[i] != t.IM[i] && this->IM[i] != '-' ) return false;
    return true;
}

bool Term::Contains( const Term& t ) const
/*
    Returns: does it fully contain t?
    Only for f-type PLA
*/
{
    unsigned int i;

    if ( t.OM.size() != this->OM.size() )
		throw RuntimeException("Term::Contains: Comparing incompatible terms");
    for ( i = 0; i < OM.size(); i++ )
        if ( this->OM[i] != '1' && t.OM[i] == '1' ) return false;
    return Contains_noOM_Check( t );
}

Term Term::Intersection( const Term& t ) const
/*
    Returns intersection of the terms
    The terms must have an intersection
    !! Ignores outputs !!
*/
{
    unsigned int i;
    Term t1;

    t1 = *this;
    for ( i = 0; i < IM.size(); i++ ) {
        if ( t1.IM[i] != t.IM[i] && t1.IM[i] != '-' && t.IM[i] != '-' ) throw RuntimeException( "Term:Intersection: no intersection" );;
        if ( t.IM[i] == '0' || t.IM[i] == '1' ) t1.IM[i] = t.IM[i];
    }
    return t1;
}

void Term::Merge( const Term& t )
/*
    Merge the terms by output matrices
    !! the terms are considered to have equal IM
    !! only for fd types
*/
{
    unsigned int i;

    for ( i = 0; i < OM.size(); i++ )
        if ( t.OM[i] == '1' ) this->OM[i] = t.OM[i];
}

bool Term::EqualIM( const Term& t ) const
	// returns true, if IMs are equal
{
	return this->IM == t.IM;
}